#! /usr/bin/env python
# Self-balancing robot simulation with visualizatio
#
# Copyright (c) 2018 Manuel Aiple
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import Tkinter as tk # for visualization
from threading import Thread # to let sim and visualization run "in parallel"
from time import clock, sleep # for wall time tracking and fake "real-time"
from math import sin, cos, pi, copysign # for trigonometric relations

class RobotSim (Thread):
    """
    Simulation thread running in fake "real-time", calculating the model
    """
    def __init__(self):
        """
        Constructor
        """
        Thread.__init__(self)
        self.reset()

    def reset(self):
        """
        Initialize all simulation variables, can be called to restart simulation.
        """
        # sampling period (in s)
        self.Ts = 1e-4
        # linear position pid controller P gain (in N/m)
        self.Kpx = 0.7
        # linear position pid controller I gain (in N/m/s)
        self.Kix = 0
        # linear position pid controller D gain (in N*s/m)
        self.Kdx = 0.9
        # linear position error (in m)
        self.xerr = 0
        # linear position integration of error (in m*s)
        self.xierr = 0
        # linear position derivative of error (in m/s)
        self.xderr = 0
        # linear position last error (to calculate xderr, in m)
        self.xerrlast = 0
        # rotational position pid controller P gain (in N/rad)
        self.Kpt = 3
        # rotational position pid controller I gain (in N/rad/s)
        self.Kit = 0
        # rotational postion pid controller D gain (in N*s/rad)
        self.Kdt = 1.2
        # rotational position error (in rad)
        self.terr = 0
        # rotational position integration of error (in rad*s)
        self.tierr = 0
        # rotational position derivative of error (in rad/s)
        self.tderr = 0
        # rotational position last error (to calculate tderr, in rad)
        self.terrlast = 0
        # gravitational constant (in m/s^2)
        self.g = 9.81
        # mass of the rod tip (in kg)
        self.mtip = 0.5
        # mass of the base (in kg)
        self.mbase = 2
        # length of the rod (in m)
        self.l = 1
        # sample index
        self.k = 0
        # linear position set point (in m)
        self.xset = 0
        # linear position set point interpolation (in m)
        self.xseti = 0
        # maximum linear position change (in m)
        self.deltaxmax = 2 * self.Ts
        # linear position (in m)
        self.x = 0
        # linear velocity (in m/s)
        self.v = 0
        # linear acceleration (in m/s^2)
        self.a = 0
        # rotational position (in rad)
        self.theta = 0
        # rotational position set point (in rad)
        self.thetaset = 0
        # rotational velocity (in rad/s)
        self.omega = 0
        # rotational acceleration (in rad/s^2)
        self.psi = 0
        # external torque (in N*m)
        self.taue = 0
        # wheel radius (in m)
        self.r = 0.3
        # floor stiffness for contact model (in N/m)
        self.Kfloor = 10000
        # floor damping for contact model (in N*s/m)
        self.Bfloor = 100
        # wheel rotation angle (in rad)
        self.alpha = self.x / (self.r)
        # generate sinusoidal linear position set point
        self.periodic = False
        # sinusoidal set point amplitude (in m)
        self.periodicAmplitude = 1.5
        # sinusoidal set point frequency (in Hz)
        self.periodicFrequency = 0.1
        # time scaling factor (< 1 faster than real-time)
        self.timeScale = 1
        # stop simulation after next iteration
        self.stopSim = False
        # simulation start time in wall time (in s)
        self.t0 = clock()
        # reset simulation at next iteration
        self.doReset = False
        # activate controller
        self.doControl = False

    def run(self):
        """
        Run the simulation model
        """
        print("Starting sim thread\n")
        # run simulation until stop requested by gui
        while self.stopSim == False:
            # reset simulation if requested by gui
            if self.doReset == True:
                self.reset()

            # calculate simulation time and do fake "real-time"
            self.k += 1
            self.t = self.k * self.Ts
            while (clock() - self.t0) < (self.t*self.timeScale):
                sleep(self.Ts/1000.)
            
            # rotational component
            self.theta += self.omega * self.Ts
            self.omega += self.psi * self.Ts
            self.taug = self.g * sin(self.theta)
            self.taua = self.a * cos(self.theta)
            # simple collision model when the tip falls on the floor
            self.ytip = self.r + self.l * cos(self.theta) - 0.04
            self.ytipdot = - self.l * sin(self.theta) * self.omega
            if self.ytip < 0:
                if self.theta > 0:
                    self.tauc = self.ytip * self.Kfloor + self.ytipdot * self.Bfloor
                else:
                    self.tauc = - (self.ytip * self.Kfloor + self.ytipdot * self.Bfloor)
                self.Ff = - self.Bfloor * self.v
            else:
                self.tauc = 0
                self.Ff = 0
            # second law of motion for rotation
            self.psi = 1/self.l * (self.taug + self.taua + self.taue + self.tauc)
            # reset external torque to obtain impulse effect
            self.taue = 0
            # linear component
            self.x += self.v * self.Ts
            self.v += self.a * self.Ts
            # sinusoidal set point generator
            if self.periodic == True:
                self.xset = self.periodicAmplitude * sin(2 * pi * self.t * self.periodicFrequency)
            # interpolator for linear set point
            deltax = 5 * (self.xset - self.xseti) * self.Ts
            if abs(deltax) < self.deltaxmax:
                self.xseti += deltax
            else:
                self.xseti += copysign(self.deltaxmax, deltax)
            # linear position pid controller
            self.xerr = self.x - self.xseti
            self.xierr += self.xerr * self.Ts
            self.xderr = (self.xerr - self.xerrlast) / self.Ts
            self.xerrlast = self.xerr
            # rotational position pid controller
            self.thetaset = 0
            self.terr =  self.theta - self.thetaset
            self.tierr += self.terr * self.Ts
            self.tderr = (self.terr - self.terrlast) / self.Ts
            self.terrlast = self.terr
            # forces from controller (actual force = set force simplification)
            self.Fx =  (self.Kpx * self.xerr + self.Kix * self.xierr + self.Kdx * self.xderr) / self.Ts
            self.Fm = - (self.Kpt * self.terr + self.Kit * self.tierr + self.Kdt * self.tderr) / self.Ts
            # forces from rotational component
            self.Fg = self.mtip * self.l * self.psi * cos(self.theta)
            self.Fc = - self.mtip * self.l * self.omega ** 2
            # second law of motion for translation
            if self.doControl == True:
                self.a = 1/(self.mbase + self.mtip) * (self.Fm + self.Fg + self.Ff + self.Fc + self.Fx)
            else:
                self.a = 1/(self.mbase + self.mtip) * (self.Fg + self.Fc + self.Ff)
            # calculate wheel rotation angle
            self.alpha = self.x / self.r

class RobotGui (Thread):
    """
    Gui thread doing the visualization
    """
    def __init__(self, sim):
        """
        Constructor
        """
        Thread.__init__(self)
        self.sim = sim
        self.canvasWidth = 1920
        self.canvasHeight = 1080
        self.screenX0 = self.canvasWidth/2
        self.screenY0 = self.canvasHeight/2
        self.screenScaleX = 400
        self.screenScaleY = -400
        self.baseRad = sim.r * self.screenScaleX
        self.tipRad = 15
        self.setPointRad = 10
        self.wheeldotRad = 5
        self.disturbanceCounter = 0
        self.disturbancePointRad = 12
        self.disturbanceOffset = 0
        self.cursorX = 0

    def screenCoords(self, x, y):
        """
        Translate world coordinates to screen coordinates
        """
        return (x * self.screenScaleX + self.screenX0, y * self.screenScaleY + self.screenY0)

    def drawLoop(self):
        """
        Draw the visualisation once and reschedule for next drawing
        """
        # delete previous elements from canvas
        items = self.canvas.find_withtag("dynamic")
        for i in items:
            self.canvas.delete(i)
        
        # calculate screen coordinates of the elements
        baseCoords = self.screenCoords(sim.x, 0)
        tipCoords = self.screenCoords(sim.x + sim.l * sin(sim.theta), sim.l * cos(sim.theta))
        xSetCoords = self.screenCoords(sim.xset, 0)
        xiSetCoords = self.screenCoords(sim.xseti, 0)
        wheeldotCoords = self.screenCoords(sim.x - 0.8*sim.r*sin(sim.alpha), -0.8*sim.r*cos(sim.alpha))
        disturbanceCoords = tipCoords
        disturbanceCoords = (disturbanceCoords[0] + self.disturbanceOffset, disturbanceCoords[1])

        # draw the elements
        self.canvas.create_line(0, self.canvasHeight/2+self.baseRad,
                            self.canvasWidth, self.canvasHeight/2+self.baseRad, width=5, tag="dynamic")
        self.canvas.create_line(baseCoords[0], baseCoords[1],
                            tipCoords[0], tipCoords[1], width=5, tag="dynamic")
        self.canvas.create_oval(baseCoords[0]-self.baseRad, baseCoords[1]-self.baseRad,
                            baseCoords[0]+self.baseRad, baseCoords[1]+self.baseRad, width=5, fill="white", tag="dynamic")
        self.canvas.create_oval(baseCoords[0]-1, baseCoords[1]-1,
                            baseCoords[0]+1, baseCoords[1]+1, fill="black", tag="dynamic")
        self.canvas.create_oval(tipCoords[0]-self.tipRad, tipCoords[1]-self.tipRad,
                            tipCoords[0]+self.tipRad, tipCoords[1]+self.tipRad, fill="black", tag="dynamic")
        self.canvas.create_oval(xSetCoords[0]-self.setPointRad, xSetCoords[1]-self.setPointRad,
                            xSetCoords[0]+self.setPointRad, xSetCoords[1]+self.setPointRad, fill="green", tag="dynamic")
        self.canvas.create_oval(xiSetCoords[0]-self.setPointRad, xiSetCoords[1]-self.setPointRad,
                            xiSetCoords[0]+self.setPointRad, xiSetCoords[1]+self.setPointRad, fill="darkgreen", tag="dynamic")
        self.canvas.create_oval(wheeldotCoords[0]-self.wheeldotRad, wheeldotCoords[1]-self.wheeldotRad,
                            wheeldotCoords[0]+self.wheeldotRad, wheeldotCoords[1]+self.wheeldotRad, fill="black", tag="dynamic")
        if self.disturbanceCounter > 0:
            self.disturbanceCounter -= 1
            self.canvas.create_oval(disturbanceCoords[0]-self.disturbancePointRad, disturbanceCoords[1]-self.disturbancePointRad,
                            disturbanceCoords[0]+self.disturbancePointRad, disturbanceCoords[1]+self.disturbancePointRad,
                            fill="blue", tag="dynamic")

        # print simulation time and info
        self.canvas.create_text(100, 15, text="wall time = %f s" % (clock()-sim.t0), tag="dynamic")
        self.canvas.create_text(100, 30, text="sim time = %f s" % (sim.t), tag="dynamic")
        self.canvas.create_text(100, 150, text="u: disturbance left\na: disturbance right\n"
                            +"n: step left\nt: step right\n"
                            +"x: go left\nv: go right\np: toggle periodic\n"
                            +"r: reset simulation\nc: toggle controller", tag="dynamic")

        # schedule redraw after 16 ms for 60 Hz update rate
        self.root.after(16, self.drawLoop)

    def keyPressed(self, event):
        """
        Key input handling function
        """
        if event.char == 'u':
            # disturbance left
            sim.taue = 3/sim.Ts
            self.disturbanceCounter = 10
            self.disturbanceOffset = 50
        elif event.char == 'a':
            # disturbance right
            sim.taue = -3/sim.Ts
            self.disturbanceCounter = 10
            self.disturbanceOffset = -50
        elif event.char == 'n':
            # move a step left
            sim.xset -= pi * sim.r / 4
        elif event.char == 't':
            # move a step right
            sim.xset += pi * sim.r / 4
        elif event.char == 'x':
            # set point to left side
            sim.xset = -1.5
        elif event.char == 'v':
            # set point to right side
            sim.xset = 1.5
        elif event.char == 'p':
            # toggle periodic on/off
            sim.periodic = not sim.periodic
        elif event.char == 'r':
            # reset simulation
            sim.doReset = True
        elif event.char == 'c':
            # toggle controller on/off
            sim.doControl = not sim.doControl

    def run(self):
        """
        Run gui thread
        """
        print("Starting gui thread\n")
        self.root = tk.Tk()
        self.canvas = tk.Canvas(self.root, width=self.canvasWidth, height=self.canvasHeight, background="white")
        self.canvas.pack()
        self.root.bind("<Key>", self.keyPressed)
        self.drawLoop()
        self.root.mainloop()
        self.sim.stopSim = True

sim = RobotSim()
gui = RobotGui(sim)

if __name__ == '__main__':
    sim.start()
    gui.start()